#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2017 Adriano Rafael Gomes <adrianorg@arg.eti.br>.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pygame
import os
import random
import math
import sys
import codecs

def clamp(val, minv, maxv):
    return minv if val < minv else maxv if val > maxv else val

class Mensagem(object):
    def __init__(self, rect, fonte, tamanho, cor, texto):
        self._rect = rect
        self._fonte = pygame.font.Font(fonte, tamanho)
        self._tamanho = tamanho
        self._cor = cor
        self._texto = u''
        self._tempo = -1
        self.definir_texto(texto)

    def _get_texto(self):
        return self._texto

    def _set_texto(self, texto):
        self.definir_texto(texto)

    texto = property(_get_texto, _set_texto)

    def definir_texto(self, texto):
        self._texto = texto
        self._atualizar_imagem()
        self._tempo = 1000

    def _atualizar_imagem(self):
        imagens = []
        linha = ''
        for palavra in self._texto.split(' '):
            if len(linha + ' ' + palavra) > 30:
                imagem = self._fonte.render(
                        linha, True, self._cor).convert_alpha()
                imagens.append(imagem)
                linha = palavra
            else:
                linha = linha + ' ' + palavra
        if len(linha) > 0:
            imagem = self._fonte.render(
                    linha, True, self._cor).convert_alpha()
            imagens.append(imagem)

        r = pygame.Rect(0, 0, 0, 0)
        for i in imagens:
            r.width = max(r.width, i.get_rect().width)
            r.height += i.get_rect().height
        imagem = pygame.Surface(r.size).convert_alpha()
        imagem.fill((255, 255, 255))
        y = 0
        for i in imagens:
            x = (r.width - i.get_rect().width) / 2
            imagem.blit(i, (x, y))
            y += i.get_rect().height
        self._imagem = imagem

    def set_tempo(self, tempo):
        self._tempo = tempo

    def update(self, dt):
        if self._tempo < 0:
            return
        self._tempo -= dt
        if self._tempo <= 0:
            self._tempo = 0

    def paint(self, surface):
        if self._tempo == 0:
            return
        pygame.draw.rect(surface, (255, 255, 255), self._rect)
        r = self._imagem.get_rect().copy()
        r.center = self._rect.center
        surface.blit(self._imagem, r)

class Camera(object):
    def __init__(self, alvo, rect_tela, rect_mundo):
        self._alvo = alvo
        self._rect_tela = rect_tela.copy()
        self._rect_mundo = rect_mundo.copy()

    def aplicar(self, rect):
        dx = clamp(self._alvo.get_rect().centerx - 300,
                0,
                self._rect_mundo.width - self._rect_tela.width)
        return rect.move(-dx, 0)

class Anim(object):
    IMAGEM = 0
    TEMPO = 1

    def __init__(self, quadros):
        self._quadros = []
        self._num_quadro_atual = 0
        self._tempo_decorrido = 0
        self._pausa = False

        self._carregar_quadros(quadros)
        self.set_quadro(0)

    def _carregar_quadros(self, quadros):
        for arquivo, tempo in quadros:
            imagem = pygame.transform.smoothscale(
                    pygame.image.load(
                        os.path.join('assets', arquivo)).convert_alpha(),
                    (50, 50))
            self._quadros.append((imagem, tempo))

    def set_quadro(self, num_quadro):
        self._num_quadro_atual = num_quadro
        self._tempo_decorrido = 0

    def proximo_quadro(self):
        num_quadro = self._num_quadro_atual + 1
        if num_quadro == len(self._quadros):
            num_quadro = 0
        self.set_quadro(num_quadro)

    def get_imagem_do_quadro_atual(self):
        return self._quadros[self._num_quadro_atual][self.IMAGEM]

    def get_tempo_do_quadro_atual(self):
        return self._quadros[self._num_quadro_atual][self.TEMPO]

    def pausar(self, pausa=True):
        self._pausa = pausa

    def get_pausa(self):
        return self._pausa

    def update(self, dt):
        if self._pausa == True:
            return
        if self._tempo_decorrido >= self.get_tempo_do_quadro_atual():
            self.proximo_quadro()
        self._tempo_decorrido += dt

class Passaro(object):
    def __init__(self, mundo):
        self._mundo = mundo
        self._velocidade_y = 0
        self._anim = Anim([('passaro-1.png', 50), ('passaro-2.png', 50)])
        self._rect = self._anim.get_imagem_do_quadro_atual().get_rect().move(
                100, 200)
        self._rastro = []
        self._tempo_rastro = 0
        self._tempo_colisao = 0
        self._pausa = True
        self._modo_automatico = False

    def pausar(self, pausa=True):
        self._pausa = pausa

    def inverter_modo_automatico(self):
        self._modo_automatico = not self._modo_automatico

    def update(self, dt):
        self._anim.update(dt)
        self._update_movimento(dt)
        self._update_rastro(dt)

    def _update_rastro(self, dt):
        self._tempo_rastro += dt
        if self._tempo_rastro > 100:
            self._tempo_rastro = 0
            self._rastro.append(self._rect.center)
            if len(self._rastro) > 10:
                self._rastro.pop(0)

    def _get_proximo_cano(self):
        rects = [ r for r in self._mundo.get_canos_rects() \
                if r.top > 0 and r.right > self._rect.right ]
        rects.sort(key=lambda r: r.left)
        if len(rects) > 0:
            return rects[0]
        else:
            return None

    def _update_movimento(self, dt):
        if self._pausa == True:
            return

        acao = {}
        acao['up']    = pygame.key.get_pressed()[pygame.K_UP] or \
                pygame.key.get_pressed()[pygame.K_SPACE]
        acao['right'] = pygame.key.get_pressed()[pygame.K_RIGHT]
        acao['left']  = pygame.key.get_pressed()[pygame.K_LEFT]

        if self._modo_automatico:
            cano = self._get_proximo_cano()
            if cano:
                dy = cano.top - self._rect.bottom
                dx = cano.left - self._rect.right
                if dy < 25 and self._velocidade_y > (dy - 100):
                    acao['up'] = True

        vx = 80
        if acao['up']:
            self._anim.pausar(False)
            self._velocidade_y = max(-250,
                    self._velocidade_y - 400 * dt / 1000.0)
        else:
            self._anim.pausar(True)
            self._velocidade_y = min(250,
                    self._velocidade_y + 300 * dt / 1000.0)
        if acao['right']:
            self._anim.pausar(False)
            vx = 240
        if acao['left']:
            self._anim.pausar(False)
            vx = 30

        if self._mundo.colisao_no_chao(self._rect):
            self._rect.bottom = self._mundo.get_chao().get_rect().top
            self._velocidade_y *= -1

        fruta, i = self._mundo.colisao_nas_frutas(self._rect)
        if i != -1:
            self._mundo.remover_fruta(i)

        cano, parte = self._mundo.colisao_nos_canos(self._rect)
        if cano:
            r = cano.get_rect()
            if parte == 'corpo':
                self._tempo_colisao = 500
            elif parte == 'topo':
                self._rect.bottom = r.top - 1
                self._velocidade_y *= -1
            elif parte == 'base':
                self._rect.top = r.bottom + 1
                self._velocidade_y = 0

        if self._tempo_colisao > 0:
            self._tempo_colisao -= dt
            vx = -100
            if self._tempo_colisao < 0:
                self._tempo_colisao = 0

        self._rect.move_ip(round(vx * dt / 1000.0),
                round(self._velocidade_y * dt / 1000.0))
        if not self._mundo.get_rect().contains(self._rect):
            self._rect.clamp_ip(self._mundo.get_rect())
            self._velocidade_y = 0

        if self._anim.get_pausa() == True:
            if self._velocidade_y > 15:
                self._anim.set_quadro(0)
            else:
                self._anim.set_quadro(1)

    def paint(self, surface, camera):
        for i, rastro in enumerate(self._rastro):
            r = pygame.Rect(rastro, (0, 0))
            c = 191 + 64 / 10.0 * i
            pygame.draw.circle(surface, (c, c, c), camera.aplicar(r).topleft, 2)

        surface.blit(self._anim.get_imagem_do_quadro_atual(),
                camera.aplicar(self._rect))

    def get_rect(self):
        return self._rect

class Imagens(object):
    def __init__(self, nomes):
        self._imagens = {}
        for nome in nomes:
            arquivo = nome + '.png'
            imagem = pygame.transform.smoothscale(
                    pygame.image.load(
                        os.path.join('assets', arquivo)).convert_alpha(),
                    (50, 50))
            self._imagens[nome] = imagem

    def get_imagem(self, nome):
        return self._imagens[nome]

    def get_imagem_qq(self):
        return random.choice(self._imagens.values())

class Fruta(object):
    def __init__(self, rect, imagem):
        self._rect = rect.copy()
        self._imagem = imagem

    def get_rect(self):
        return self._rect

    def paint(self, surface, camera):
        surface.blit(self._imagem, camera.aplicar(self._rect))

class Cano(object):
    def __init__(self, rect):
        self._rect = rect.copy()

    def get_rect(self):
        return self._rect

    def paint(self, surface, camera):
        corpo = self._rect.copy()
        topo  = self._rect.copy()
        base  = self._rect.copy()

        corpo.inflate_ip(-10, 0)
        topo.height = 25
        base.height = 25
        base.bottom = corpo.bottom

        self._paint_rect(surface, camera, corpo)
        self._paint_rect(surface, camera, topo)
        self._paint_rect(surface, camera, base)

    def _paint_rect(self, surface, camera, rect):
        CORPO  = (63, 191, 63)
        BRILHO = (191, 255, 191)
        BORDA  = (0, 127, 0)

        corpo  = rect.copy()
        brilho = rect.copy()
        borda  = rect.copy()

        brilho.left += 10
        brilho.width = 10

        pygame.draw.rect(surface, CORPO,  camera.aplicar(corpo),  0)
        pygame.draw.rect(surface, BRILHO, camera.aplicar(brilho), 0)
        pygame.draw.rect(surface, BORDA,  camera.aplicar(borda),  3)

class Chao(object):
    def __init__(self, rect_mundo):
        self._imagem = pygame.transform.smoothscale(
                pygame.image.load(
                    os.path.join('assets', 'chao.png')).convert_alpha(),
                (204, 104))
        self._rect = self._imagem.get_rect()
        self._rect.bottom = rect_mundo.bottom
        self._rect.width = rect_mundo.width

    def get_rect(self):
        return self._rect

    def paint(self, surface, camera):
        r = self._imagem.get_rect()
        r.top = self._rect.top
        for x in range(
                self._rect.left,
                self._rect.right,
                self._imagem.get_rect().width):
            r.left = x
            surface.blit(self._imagem, camera.aplicar(r))

class Mundo(object):
    def __init__(self, rect_tela, nivel):
        self._rect = None
        self._chao = None
        self._canos = []
        self._canos_rects_cache = []
        self._frutas = []
        self._frutas_rects_cache = []
        self._frutas_imagens = Imagens(
                ('banana', 'cereja', 'laranja', 'maca', 'melancia', 'pera'))

        qtd_canos = nivel
        dx = max(300, 200 + int(400 / math.sqrt(qtd_canos)))
        self._rect = rect_tela.copy()
        self._rect.width = max(rect_tela.width, 400 + qtd_canos * dx)
        self._chao = Chao(self._rect)
        self._criar_canos_e_frutas(nivel, qtd_canos, dx)

    def _criar_canos_e_frutas(self, nivel, qtd_canos, dx):
        canos = []
        canos_rects_cache = []
        frutas = []
        frutas_rects_cache = []
        y = 0
        for i in range(qtd_canos):
            y1 = self._rect.top + 100
            y2 = self._chao.get_rect().top - 100
            f = min(10, nivel) / 10.0
            dy = int(random.random() * (y2 - y1) * f) * random.choice([-1, 1])
            y = (y2 - y1 + dy) / 2 + y1
            x = int(400 + i * dx + random.randint(-50, 50))
            r1 = pygame.Rect(0, 0, 50, self._rect.height)
            r2 = pygame.Rect(0, 0, 50, self._rect.height)
            r3 = pygame.Rect(0, 0, 50, 50)
            r1.centerx = x
            r1.bottom = y - 50
            r2.centerx = x
            r2.top = y + 50
            r3.center = (x, y)
            canos.append(Cano(r1))
            canos.append(Cano(r2))
            frutas.append(Fruta(r3, self._frutas_imagens.get_imagem_qq()))
        self._canos = canos
        self._canos_rects_cache = [ cano.get_rect() for cano in self._canos ]
        self._frutas = frutas
        self._frutas_rects_cache = [ fruta.get_rect() for fruta in self._frutas ]

    def get_rect(self):
        return self._rect

    def paint(self, surface, camera):
        for cano in self._canos:
            cano.paint(surface, camera)
        for fruta in self._frutas:
            fruta.paint(surface, camera)
        self._chao.paint(surface, camera)

    def colisao_nos_canos(self, rect):
        i = rect.collidelist(self._canos_rects_cache)
        if i == -1:
            return None, None
        else:
            parte = None
            corpo = self._canos[i].get_rect().copy()
            topo  = self._canos[i].get_rect().copy()
            base  = self._canos[i].get_rect().copy()

            corpo.inflate_ip(0, -50)
            topo.height = 25
            base.height = 25
            base.top = corpo.bottom

            if rect.colliderect(corpo):
                parte = 'corpo'
            elif rect.colliderect(topo):
                parte = 'topo'
            elif rect.colliderect(base):
                parte = 'base'
            return self._canos[i], parte

    def colisao_no_chao(self, rect):
        return rect.bottom > self._chao.get_rect().top

    def colisao_nas_frutas(self, rect):
        i = rect.collidelist(self._frutas_rects_cache)
        if i == -1:
            return None, i
        else:
            return self._frutas[i], i

    def get_chao(self):
        return self._chao

    def get_canos_rects(self):
        return self._canos_rects_cache

    def remover_fruta(self, i):
        self._frutas.pop(i)
        self._frutas_rects_cache.pop(i)

def exibir_mensagem(mensagem, nivel):
    with codecs.open(os.path.join('assets', 'mensagens.txt'), 'r',
            encoding='utf-8') as f:
        linhas = f.readlines()
        n = nivel % len(linhas) - 1
        linha = linhas[n].rstrip()
        mensagem.definir_texto(u'Nível ' + unicode(nivel) + u': ' + linha)
        mensagem.set_tempo(-1)

def main(nivel):
    random.seed()
    pygame.init()
    clock = pygame.time.Clock()

    tela = pygame.display.set_mode((800, 600))
    pygame.display.set_caption('Passarinho')
    mundo = Mundo(tela.get_rect(), nivel)
    passaro = Passaro(mundo)
    camera = Camera(passaro, tela.get_rect(), mundo.get_rect())
    mensagem = Mensagem(
            tela.get_rect().inflate(-400, -400).move(0, -50),
            None, 30, (0, 0, 0), u'')
    exibir_mensagem(mensagem, nivel)

    pygame.mixer.music.load(os.path.join('assets', 'musica.ogg'))
    pygame.mixer.music.play(-1)
    musica_tocando = True

    frame = 1
    continuar = True
    passaro.pausar(True)
    while continuar == True:
        for e in pygame.event.get():
            if e.type == pygame.QUIT:
                continuar = False
                break
            elif e.type == pygame.KEYDOWN:
                if e.key == pygame.K_ESCAPE:
                    continuar = False
                elif e.key == pygame.K_m:
                    if musica_tocando:
                        pygame.mixer.music.pause()
                    else:
                        pygame.mixer.music.unpause()
                    musica_tocando = not musica_tocando
                elif e.key == pygame.K_a:
                    passaro.inverter_modo_automatico()
                else:
                    mensagem.set_tempo(0)
                    passaro.pausar(False)

        dt = clock.tick(30)
        passaro.update(dt)
        mensagem.update(dt)
        if passaro.get_rect().left > mundo.get_rect().right - 100:
            passaro.pausar(True)
            nivel += 1
            exibir_mensagem(mensagem, nivel)
            mundo = Mundo(tela.get_rect(), nivel)
            passaro = Passaro(mundo)
            camera = Camera(passaro, tela.get_rect(), mundo.get_rect())
        tela.fill((127, 191, 255))
        mundo.paint(tela, camera)
        passaro.paint(tela, camera)
        mensagem.paint(tela)
        pygame.display.flip()
        #pygame.image.save(tela, 'img_%06d.png' % frame)
        frame += 1

    pygame.quit()

if __name__ == '__main__':
    if len(sys.argv) > 1:
        nivel = int(sys.argv[1])
    else:
        nivel = 1
    main(nivel)
